/* Funcoes graficas basicas para operar video.

Assembler AT&T + DJGPP:

- se gravar em variavel de memoria passada pelos parametros
  por "memory" no 3 :;
- __volatile__ permite ao compilador reorganizar o codigo
*/



#include <go32.h>
#include <string.h>   // necessario para acessar sys/movedata.h
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <pc.h>

#include "defs.h"

#include "gr.h"



// calcula incremento de x (inverso da declividade)
#define INCREMENTOX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* construtor - Aloca memoria, inicia modo grafico, programa palette,...
**********************************************************/
CGr::CGr ()
{
	// tenta alocar memoria para memoria virtual 0, se nao conseguir sai
	if ( ( m_pyVr [ 0 ] = new BYTE [ TAMVR ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

	// tenta alocar memoria para memoria virtual 1, se nao conseguir sai
	if ( ( m_pyVr [ 1 ] = new BYTE [ TAMVR ] ) == NULL ) {
   	delete m_pyVr [ 0 ];
      m_pyVr [ 0 ] == NULL;
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

	// inicia com a memoria virtual 0 ativa
	m_nVrAtiva = 0;
	mAtrVrCor ( 0 );

	// entra em modo grafico
	mAtrVdModo ( MCGA );

	// faz com que a proporcao entre X e Y dos pixels seja 1
	outportb ( 0x3C2, 0xE3 );

	// programa pallete
   // pallete eh dividida em 8 blocos de tons de cores diferentes
	for ( INT ncBase = 0; ncBase < 8; ncBase ++ ) {
	  	BYTE
			yTom = ILUMAMB;     // o numero eh o tom mais escuro usado
		// programa cada bloco com variacoes de tom da mesma cor
		for ( INT ncC = ncBase * NUMCOR; ncC < ( ncBase * NUMCOR + NUMCOR ); ncC ++ ) {

         // seleciona cor base
      	switch ( ncBase ) {
         	case 0:    // 0 - indice da pallete
					mAtrPalCor ( ncC, yTom, yTom, yTom );
               break;
         	case 1:    // 32
					mAtrPalCor ( ncC, yTom, yTom, 5 );
               break;
         	case 2:    // 64
					mAtrPalCor ( ncC, 5, 5, yTom );
               break;
         	case 3:    // 96
					mAtrPalCor ( ncC, 5, yTom, yTom );
               break;
         	case 4:    // 128
					mAtrPalCor ( ncC, 5, yTom, 5 );
               break;
				case 5:    // 160
					mAtrPalCor ( ncC, 20, yTom, 20 );
					break;
         	case 6:    // 192
					mAtrPalCor ( ncC, yTom - 20, yTom - 20, yTom - 20 );
					break;
         	case 7:    // 224
					mAtrPalCor ( ncC, yTom, yTom / 2, yTom );
               break;
         }
         yTom ++;
      }
   }
	mAtrPalCor ( 0, 0, 0, 0 );
}



/* destrutor - Volta ao modo texto, desaloca memoria
**********************************************************/
CGr::~CGr ()
{
	// entra em modo texto
	mAtrVdModo ( TEXTO );

	// desaloca
	delete m_pyVr [ 0 ];
   if ( m_pyVr [ 1 ] != NULL )
		delete m_pyVr [ 1 ];
}



/* Altera o modo de exibicao do video.
**********************************************************/
void CGr::mAtrVdModo ( UINT16 uModo )
{
	REGS
		stRegsE, stRegsS;      // registradores

	stRegsE.x.ax = uModo;
	int86 ( 0x10, &stRegsE, &stRegsS );
}



/* Atribui uma quantidade de RGB para uma cor (entrada) da palette
**********************************************************/
void CGr::mAtrPalCor ( BYTE yCor, BYTE yR, BYTE yG, BYTE yB )
{
	outportb ( 0x03C8, yCor );
	outportb ( 0x03C9, yR );
	outportb ( 0x03C9, yG );
	outportb ( 0x03C9, yB );
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
**********************************************************/
void CGr::mVdSincRetraco ()
{
	__asm__ __volatile__ ("

			 movw  $0x03DA, %%dx
		L1: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jnz   L1

		L2: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    L2
		"
		:
		:
		: "%dx", "%al"
	);
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
  durante a espera zera a pagina de video virtual nao ativa.
__asm__ __volatile__ ("

		movl  %0,      %%edi   edi com endereco inicial a ser zerado
		movl  $16000,  %%ecx   ecx recebe o numero de interacoes (16000)
		xorl  %%ebx,   %%ebx   ebx recebe 0 (valor a ser escrito)

		movw  $0x03DA, %%dx    parte do "WaintRetrace"

  A1: movl  %%ebx,  (%%edi)  escreve no endereco apontado por edi um zero de 4 bytes (conteudo de ebx)
		addl  $4,      %%edi   soma 4 bytes no edi, proximos 4 bytes a serem escritos
		subl  $1,      %%ecx   subtrai 1 de ecx, menos uma interacao
		cmpl  $0,      %%ecx   se nao existe mais memoria para zerar contador de interacoes (ecx) chegou a zero nao zera mais a memoria
		jz    A2               se ecx == 0 desvia para parte onde soh faz primeiro teste do "WaitRetrace"

		inb   %%dx,    %%al    primeiro teste do "WaitRetrace"
		andb  $0x8,    %%al
		jz    A1               se o teste nao teve sucesso volta

  A4: movl  %%ebx,  (%%edi)  mesma escrita em endereco que em A1, soh depois de passar primeiro teste do "WaitRetrace"
		addl  $4,      %%edi
		subl  $1,      %%ecx
		cmpl  $0,      %%ecx
		jz    A3               se terminou de zerar memoria desvia para parte sem escrita em memoria (soh teste)

		inb   %%dx,    %%al    segundo teste do "WaitRetrace"
		andb  $0x8,    %%al
		jz    A4               se falhou teste volta

		jmp   AFimAtr          se acertou teste termina escrita na memoria antes de sair

  A2: inb   %%dx,    %%al    primeiro teste do "WaitRetrace" sem zerar memoria
		andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
		jnz   A2

  A3: inb   %%dx,    %%al    segundo teste do "WaitRetrace" sem zerar memoria
		andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
		jz    A3

		jmp AFim               termina sem passar por termino de escrita em memoria (logo abaixo), escrita jah acabou

  AFimAtr:
		movl  %%ebx,  (%%edi)  termina de escrever a memoria (apohs ter passado pelos 2 testes de "WaitRetrace")
		addl  $4,      %%edi
		subl  $1,      %%ecx
		cmpl  $0,      %%ecx
		jnz   AFimAtr

  AFim:

**********************************************************/
void CGr::mVrSincRetraco ()
{
   __asm__ __volatile__ ("

			 movl  %0,      %%edi
			 movl  $16000,  %%ecx
			 xorl  %%ebx,   %%ebx

			 movw  $0x03DA, %%dx

	   A1: movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jz    A2

			 inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A1

	   A4: movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jz    A3

			 inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A4

			 jmp   AFimAtr

	   A2: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jnz   A2

	   A3: inb   %%dx,    %%al
			 andb  $0x8,    %%al
			 jz    A3

			 jmp   AFim

	   AFimAtr:
			 movl  %%ebx,  (%%edi)
			 addl  $4,      %%edi
			 subl  $1,      %%ecx
			 cmpl  $0,      %%ecx
			 jnz   AFimAtr

	   AFim:
	   "
	   :
	   : "g" ( m_pyVr [ ! m_nVrAtiva ] )
	   : "%dx", "%al", "%edi", "%ecx", "%eax", "%ebx"
   );
}



/* Atribui toda a tela com uma soh cor. Sincroniza com retraco.
**********************************************************/
void CGr::mAtrVdCorSinc ( BYTE yCor )
{
	BYTE
		ayCores [ 4 ];
	ayCores [ 0 ] = ayCores [ 1 ] = ayCores [ 2 ] = ayCores [ 3 ] = yCor;

	mVdSincRetraco ();

	for ( INT ncTrans = 0; ncTrans < TAMVR; ncTrans += 4 )
		_dosmemputl ( (void*) ayCores, 1, 0xA0000 + ncTrans );
}



/* Marca um pixel na tela
**********************************************************/
void CGr::mAtrVdPixel ( INT nX, INT nY, BYTE yCor )
{
	dosmemput ( (void*) &yCor, 1, 0xA0000 + nX + ( nY * RESX ) );
}



/* Atribui toda tela virtual com uma soh cor.
**********************************************************/
void CGr::mAtrVrCor ( BYTE yCor )
{
	memset ( (void*) m_pyVr [ m_nVrAtiva ], yCor, TAMVR );
}



/* Marca um pixel na memoria virtual
**********************************************************/
void CGr::mAtrVrPixel ( INT nX, INT nY, BYTE yCor )
{
	memset ( (void*) m_pyVr [ m_nVrAtiva ] + nX + ( nY * RESX ), yCor, 1 );
}



/* Transfere a memoria virtual para a tela, com sincronizacao
 de retraco. Ativa tela virtual inativa e desativa a ativa.
**********************************************************/
void CGr::mAtrVdVrSinc ()
{
	mVrSincRetraco ();

	// move dados de 4 em 4 bytes
	_movedatal ( _my_ds (), (INT) m_pyVr [ m_nVrAtiva ], _dos_ds, 0xA0000, TAMVR4 );

	// altera memoria virtual de video ativa para a que estava inativa e limpa
	m_nVrAtiva = ! m_nVrAtiva;
}



/* 1 - Desenha um poligono na memoria virtual de video usando
o incremento de x para calcular posicao de x.
Para garantir desenho de poligino convexo os vertices
adjacentes devem ocupar posicoes contiguas nos parametros
passado para essa funcao. Ou seja, visualmente, devem ser definidos
em ordem horaria ou anti-horaria. Exemplo:

								(x4, y4)
								*                -
					        / \                 loop
							 /   \                    |
				(x1, y1) *     \                   |
							 \		 * (x3, y3)         | vai calculando posicoes de x
					        \	/                   | entre retas horizontais e
								\ /                    V retas das bordas do poligono
								 *               -
								 (x2, y2)

Equacao das retas:
Y = aX + b, onde
            a eh declividade
            b eh deslocamento da reta no eixo y
Equacao usada nas interseccoes (x eh calculado definido quanto x varia a cada
linha e somando a um valor inicial de x):
X = ( deltaY * incrementoX ) + Xvertice, onde
     		   incrementoX eh 1 / a, quanto x varia por linha
		   	deltaY eh a distancia entre o Y de um vertice da reta e Y do loop
            Xvertice eh o X do vertice usado no calculo de deltaY

Os vertices sao numerados de 1 a 4 correspondendo a sua ordem nos parametros.
Sao chamados vertices superior ou inferior os vertices dos extremos
considerando y, no exemplo sao (x4, y4) e (x2, y2), respectivamente.

Parte que desenha linha horizontal:
// desenha linha horizontal, testa para nao desenhar por cima
__asm__ __volatile__ ("

	   movl  %0,      %%edi   edi recebe a coordenada inicial (endereco inicial a ser zerado)
	   movl  %1,      %%ecx   ecx recebe a coordenada final (ultimo endereco a ser zerado)
	   movb  %2,      %%bh    bh recebe a cor (valor a ser escrito)

  Des:cmpb  $0,     (%%edi)  verifica se jah foi desenhado no byte (endereco apontado por edi tem valor == 0)
      jne   Inc              se jah foi desenhado nao marca ponto e passa para proxima coluna ((%%edi) != 0)
      movb  %%bh,   (%%edi)  marca ponto (escreve no endereco apontado por edi a cor - o conteudo de ebx)
  Inc:cmpl  %%edi,   %%ecx   verifica se terminou de tracar a linha (endereco atual igual ao final)
	   je    Para             se terminou (edi == ecx) vai para fim
      incl  %%edi            vai para a proxima coluna (soma 1 byte no edi)
      jmp   Des              continua a desenhar
  Para:
	   "
	   :
	   : "g" ( uEndLinha + nXIni ), "g" ( uEndLinha + nXFin ), "g" ( yCor )
	   : "%bh", "%ecx", "%edi"
   );
**********************************************************/
void CGr::mAtrVrPol2D ( INT nX1, INT nY1, INT nX2, INT nY2, INT nX3, INT nY3,
							   INT nX4, INT nY4, BYTE yCor )
{
   // recorta limites extremos de y
	// define valores dos extremos em y (maior e menor y)
   INT16
   	nYSup = nY1,
      nYInf = nY1;
   // testa y2
	if ( nYSup > nY2 )
		nYSup = nY2;
	else
		if ( nYInf < nY2 )
			nYInf = nY2;
   // testa y3
	if ( nYSup > nY3 )
		nYSup = nY3;
	else
		if ( nYInf < nY3 )
			nYInf = nY3;
   // testa y4
	if ( nYSup > nY4 )
		nYSup = nY4;
	else
		if ( nYInf < nY4 )
			nYInf = nY4;
   // testa se extremos estao fora dos limites da tela, se sim define dentro
	if ( nYSup < MINY )
		nYSup = MINY;
	if ( nYInf > MAXY )
		nYInf = MAXY;

   // define se eh possivel calcular incrementos de x, se sim calcula
   float
   	fIncX12 = 0,
   	fIncX14 = 0,
   	fIncX23 = 0,
   	fIncX34 = 0;
   BOOL
   	bIncValido12 = TRUE,
   	bIncValido14 = TRUE,
   	bIncValido23 = TRUE,
   	bIncValido34 = TRUE;
   // sao calculados valores apenas entre pontos que nao estao em linhas ==
   if ( nY1 != nY2 )
   	fIncX12 = INCREMENTOX ( nX1, nY1, nX2, nY2 );
   else
      bIncValido12 = FALSE;
   if ( nY1 != nY4 )
   	fIncX14 = INCREMENTOX ( nX1, nY1, nX4, nY4 );
   else
      bIncValido14 = FALSE;
   if ( nY2 != nY3 )
   	fIncX23 = INCREMENTOX ( nX2, nY2, nX3, nY3 );
   else
      bIncValido23 = FALSE;
   if ( nY3 != nY4 )
   	fIncX34 = INCREMENTOX ( nX3, nY3, nX4, nY4 );
   else
      bIncValido34 = FALSE;

   // calcula endereco da primeira linha de video que serah desenhada
   UINT
      uEndLinha = (UINT) m_pyVr [ m_nVrAtiva ] + RESX * nYSup;

   // varre poligono do extremo superior ateh o extremo inferior
   INT16
   	nXIni,    // deve possuir apenas x inicial da linha horizontal
      nXFin,    // deve possuir apenas x final da linha horizontal
      nX;       // valor x da interceccao entre duas retas
	for ( INT16 ncY = nYSup; ncY <= nYInf; ncY ++ ) {

      // valores extremos iniciais de x (inicio e fim da linha horizontal)
   	nXIni = 32767;
   	nXFin = - 32768;

      // se estao sendo definidas posicoes de x entre vertices 1 e 4
      if ( bIncValido14 &&
		     ( ncY >= nY1 && ncY <= nY4 ) ||
		     ( ncY <= nY1 && ncY >= nY4 ) ) {

			nX = (INT16) ( (float) ( ncY - nY4 ) * fIncX14 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 1 e 2
      if ( bIncValido12 &&
		     ( ncY >= nY1 && ncY <= nY2 ) ||
		     ( ncY <= nY1 && ncY >= nY2 ) ) {

			nX = (INT16) ( (float) ( ncY - nY2 ) * fIncX12 ) + nX2;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 2 e 3
      if ( bIncValido23 &&
		     ( ncY >= nY2 && ncY <= nY3 ) ||
		     ( ncY <= nY2 && ncY >= nY3 ) ){

			nX = (INT16) ( (float) ( ncY - nY3 ) * fIncX23 ) + nX3;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 3 e 4
      if ( bIncValido34 &&
		     ( ncY >= nY3 && ncY <= nY4 ) ||
		     ( ncY <= nY3 && ncY >= nY4 ) ) {

			nX = (INT16) ( (float) ( ncY - nY4 ) * fIncX34 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // Recorta valores extremos da linha horizontal (maior e menor x),
		// corrige valores dos extremos em x se estiverem fora da tela.
      // verifica inicio
      if ( nXIni < MINX )
      	nXIni = MINX;
      else
         // inicio pode ser muito a direita: linha toda fora da tela
      	if ( nXIni > MAXX ) {
         	// calcula endereco da proxima linha de video
			   uEndLinha += RESX;
            // nao desenha
         	continue;
         }
      // verifica fim
      if ( nXFin > MAXX )
      	nXFin = MAXX;
      else
         // fim pode ser muito a esquerda: linha toda fora da tela
      	if ( nXFin < MINX ) {
         	// calcula endereco da proxima linha de video
			   uEndLinha += RESX;
            // nao desenha
         	continue;
         }

      // desenha linha horizontal, testa para nao desenhar por cima
      __asm__ __volatile__ ("

	         movl  %0,      %%edi
	         movl  %1,      %%ecx
	         movb  %2,      %%bh

        Des:cmpb  $0,     (%%edi)
            jne   Inc
            movb  %%bh,   (%%edi)
	     Inc:cmpl  %%ecx,   %%edi
	         je    Para
            incl  %%edi
            jmp   Des
        Para:
	         "
	         :
	         : "g" ( uEndLinha + nXIni ), "g" ( uEndLinha + nXFin ), "g" ( yCor )
	         : "%bh", "%ecx", "%edi"
      );

      // calcula endereco da proxima linha de video (pode nao ser desenhada)
      uEndLinha += RESX;
	}
}
