/* Solido formado por matriz de vertices e poligonos.
Vertices sao compostos por 3 componentes: x,y,z.
Poligonos sao compostos por 4 vertices (pode haver sobreposicao de vertices).
Os dados de cada solido estao em um arquivo de solido (*.S).
Somente pode haver 1 definicao de solido por objeto de CSolido (impossivel
redefinir). */



#include <stdio.h>       // acesso a NULL
#include <math.h>

#include "defs.h"

#include "arquivo.h"

#include "solido.h"



/* construtor
**********************************************************/
CSolido::CSolido ()
{
   m_uMaxVertices = 0;
   m_uNumVertices = 0;
	m_uMaxPoligonos = 0;
   m_uNumPoligonos = 0;
	m_panX = m_panY = m_panZ = m_panD = NULL;
	m_pauV1 = m_pauV2 = m_pauV3 = m_pauV4 = m_pauP = NULL;
	m_pauVNC1 = m_pauVNC2 = m_pauVNC3 = NULL;
   m_payCor = NULL;
}



/* construtor - cria solido compativel com outro solido
**********************************************************/
CSolido::CSolido ( CSolido &oS )
{
   mCriaCompativel ( oS );
}



/* construtor - carrega arquivo de objeto arquivo
**********************************************************/
CSolido::CSolido ( char* szArq )
{
   CArquivo
   	oArq ( szArq, *this );
}



/* construtor - carrega arquivo de objeto arquivo externo
**********************************************************/
CSolido::CSolido ( char* szArq, CArquivo& oArq )
{
	oArq.mRetObjeto ( szArq, *this );
}



/* destrutor - desaloca memoria
**********************************************************/
CSolido::~CSolido ()
{
	if ( m_panX != NULL )
   	delete m_panX;
	if ( m_panY != NULL )
   	delete m_panY;
	if ( m_panZ != NULL )
   	delete m_panZ;

	if ( m_panD != NULL )
   	delete m_panD;
	if ( m_pauP != NULL )
   	delete m_pauP;

	if ( m_pauV1 != NULL )
   	delete m_pauV1;
	if ( m_pauV2 != NULL )
   	delete m_pauV2;
	if ( m_pauV3 != NULL )
   	delete m_pauV3;
	if ( m_pauV4 != NULL )
   	delete m_pauV4;

	if ( m_pauVNC1 != NULL )
   	delete m_pauVNC1;
	if ( m_pauVNC2 != NULL )
   	delete m_pauVNC2;
	if ( m_pauVNC3 != NULL )
   	delete m_pauVNC3;

	if ( m_payCor != NULL )
   	delete m_payCor;
}



/* Aloca memoria. Deve ser chamado antes de iniciar atribuicao de dados.
**********************************************************/
void CSolido::mAlocaMemoria ( UINT16 uMaxVert, UINT16 uMaxPol )
{
   // define dimensoes
	m_uMaxVertices = uMaxVert;
   m_uNumVertices = 0;
	m_uMaxPoligonos = uMaxPol;
   m_uNumPoligonos = 0;
	m_panX = m_panY = m_panZ = m_panD = NULL;
   m_payCor = NULL;

	// tenta alocar memoria para memoria para vertices, se nao conseguir sai
	if ( ( m_panX = new INT16 [ m_uMaxVertices + m_uMaxPoligonos * 2 ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_panY = new INT16 [ m_uMaxVertices + m_uMaxPoligonos * 2 ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_panZ = new INT16 [ m_uMaxVertices + m_uMaxPoligonos * 2 ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

   // tenta alocar memoria para vertices de poligonos, se nao conseguir sai
	if ( ( m_pauV1 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauV2 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauV3 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauV4 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

   // tenta alocar memoria para orenacao
	if ( ( m_panD = new INT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauP = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

   // tenta alocar memoria para vertices nao colineares dos poligonos
	if ( ( m_pauVNC1 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauVNC2 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
	if ( ( m_pauVNC3 = new UINT16 [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}

   // tenta alocar memoria para tabela de cores
	if ( ( m_payCor = new BYTE [ m_uMaxPoligonos ] ) == NULL ) {
		printf ( MSG_SEMMEMORIA );
		exit ( 1 );
	}
}



/* Cria solido compativel com outro solido. Deve ser chamado
uma vez antes de iniciar transformacoes em um solido.
**********************************************************/
void CSolido::mCriaCompativel ( CSolido &oS )
{
   // aloca mesma quantidade de memoria
	mAlocaMemoria ( oS.mRetMaxVertices (), oS.mRetMaxPoligonos () );

   // copia definicoes de vertices (usado apenas para calcular cores)
	memcpy ( m_panX, oS.m_panX, oS.mRetMaxVertices () * sizeof ( INT16 ) );
	memcpy ( m_panY, oS.m_panY, oS.mRetMaxVertices () * sizeof ( INT16 ) );
	memcpy ( m_panZ, oS.m_panZ, oS.mRetMaxVertices () * sizeof ( INT16 ) );

   // copia definicoes de poligonos
	memcpy ( m_pauV1, oS.m_pauV1, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );
	memcpy ( m_pauV2, oS.m_pauV2, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );
	memcpy ( m_pauV3, oS.m_pauV3, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );
	memcpy ( m_pauV4, oS.m_pauV4, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );

   // copia definicoes de vertices nao colineares
	memcpy ( m_pauVNC1, oS.m_pauVNC1, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );
	memcpy ( m_pauVNC2, oS.m_pauVNC2, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );
	memcpy ( m_pauVNC3, oS.m_pauVNC3, oS.mRetMaxPoligonos () * sizeof ( UINT16 ) );

   // copia definicoes de cores de poligonos
   memcpy ( m_payCor, oS.m_payCor, oS.mRetMaxPoligonos () * sizeof ( BYTE ) );

   // define numero de poligonos do solido
   m_uNumPoligonos = oS.m_uNumPoligonos;

   // define poligonos a serem ordenados
   mOrdenarPols ();
}



/* Inclui um vertice na matriz de vertices
**********************************************************/
void CSolido::mAtrVertice ( INT16 nX, INT16 nY, INT16 nZ )
{
	* ( m_panX + m_uNumVertices ) = nX;
	* ( m_panY + m_uNumVertices ) = nY;
	* ( m_panZ + m_uNumVertices ) = nZ;

   // verifica esgotamento do espaco para vertices
   if ( m_uNumVertices < m_uMaxVertices )
   	m_uNumVertices ++;
}



/* Inclui um poligono no solido
**********************************************************/
void CSolido::mAtrPoligono ( UINT16 uV1, UINT16 uV2, UINT16 uV3, UINT16 uV4,
                             BYTE yCor )
{
   // inclui vertices
	m_pauV1 [ m_uNumPoligonos ] = uV1;
	m_pauV2 [ m_uNumPoligonos ] = uV2;
	m_pauV3 [ m_uNumPoligonos ] = uV3;
	m_pauV4 [ m_uNumPoligonos ] = uV4;

   // Define trio de vertices que serao usados para obter vetor normal,
	// (nao devem ser colineares, mas eh soh testado se nao estao sobrepostos)
   // Verifica se melhor vertice para origem dos vetores paralelos ao
	// poligono eh o vertice 1 ou 3:
   //
   //        4---1      |41| > |32| e |12| > |43| => 1 senao 3
   //        |   |
   //        3---2
   //
   // Exclui vertice de origem que tem a menor distancia entre seus
	// adjacentes correspondentes.
   // Os vertices devem estar em ordem horaria: se origem for 1: 1 - 2 - 4;
   // se origem for 3: 3 - 2 - 4. O vertice de orgiem deve ser o primeiro!
   // calcula distancias entre vertices
   float
   	fDist12 = MODULO ( m_panX [ uV2 ] - m_panX [ uV1 ],
      						 m_panY [ uV2 ] - m_panY [ uV1 ],
                         m_panZ [ uV2 ] - m_panZ [ uV1 ] ),
   	fDist32 = MODULO ( m_panX [ uV2 ] - m_panX [ uV3 ],
      						 m_panY [ uV2 ] - m_panY [ uV3 ],
                         m_panZ [ uV2 ] - m_panZ [ uV3 ] ),
   	fDist43 = MODULO ( m_panX [ uV3 ] - m_panX [ uV4 ],
      						 m_panY [ uV3 ] - m_panY [ uV4 ],
                         m_panZ [ uV3 ] - m_panZ [ uV4 ] ),
   	fDist41 = MODULO ( m_panX [ uV1 ] - m_panX [ uV4 ],
      						 m_panY [ uV1 ] - m_panY [ uV4 ],
                         m_panZ [ uV1 ] - m_panZ [ uV4 ] );
   // compara distancias
   if ( fDist41 > fDist32 && fDist12 > fDist43 )
   	// melhor vertice eh o 1
      // verifica sobreposicoes, se nao tiver sobreposicao usa o vertice 1
	   // verifica se os vertices 1 e 2 estao sobrepostos, se nao usa 1 e 2
	   if ( m_panX [ uV1 ] != m_panX [ uV2 ] ||
		     m_panY [ uV1 ] != m_panY [ uV2 ] ||
		     m_panZ [ uV1 ] != m_panZ [ uV2 ] ) {

         m_pauVNC1 [ m_uNumPoligonos ] = uV1;
         m_pauVNC2 [ m_uNumPoligonos ] = uV2;

		   // verifica se os vertices 1 e 4 estao sobrepostos, se nao usa 1 e 4
		   if ( m_panX [ uV1 ] != m_panX [ uV4 ] ||
			     m_panY [ uV1 ] != m_panY [ uV4 ] ||
			     m_panZ [ uV1 ] != m_panZ [ uV4 ] )

	         m_pauVNC3 [ m_uNumPoligonos ] = uV4;

		   // se demais vertices estavam sobrepostos: 1 e 3
		   else

	         m_pauVNC3 [ m_uNumPoligonos ] = uV3;

	   }
	   else {

         // se vertices 1 e 2 estao sobrepostos usa os que sobraram: 1, 3, 4
         m_pauVNC1 [ m_uNumPoligonos ] = uV1;
  	      m_pauVNC2 [ m_uNumPoligonos ] = uV3;
         m_pauVNC3 [ m_uNumPoligonos ] = uV4;
      }
   else
   	// melhor vertice eh o 3
      // verifica sobreposicoes, se nao tiver sobreposicao usa o vertice 3
	   // verifica se os vertices 3 e 4 estao sobrepostos, se nao usa 3 e 4
	   if ( m_panX [ uV3 ] != m_panX [ uV4 ] ||
		     m_panY [ uV3 ] != m_panY [ uV4 ] ||
		     m_panZ [ uV3 ] != m_panZ [ uV4 ] ) {

         m_pauVNC1 [ m_uNumPoligonos ] = uV3;
         m_pauVNC2 [ m_uNumPoligonos ] = uV4;

		   // verifica se os vertices 3 e 2 estao sobrepostos, se nao usa 3 e 2
		   if ( m_panX [ uV3 ] != m_panX [ uV2 ] ||
			     m_panY [ uV3 ] != m_panY [ uV2 ] ||
			     m_panZ [ uV3 ] != m_panZ [ uV2 ] )

	         m_pauVNC3 [ m_uNumPoligonos ] = uV2;

		   // se demais vertices estavam sobrepostos: 3 e 1
		   else

	         m_pauVNC3 [ m_uNumPoligonos ] = uV1;

	   }
	   else {

         // se vertices 3 e 2 estao sobrepostos usa os que sobraram: 3, 1, 4
         m_pauVNC1 [ m_uNumPoligonos ] = uV3;
  	      m_pauVNC2 [ m_uNumPoligonos ] = uV4;
         m_pauVNC3 [ m_uNumPoligonos ] = uV1;
      }

   // define cor basica original do poligono
   m_payCor [ m_uNumPoligonos ] = yCor;

   // verifica esgotamento do espaco para poligonos
   if ( m_uNumPoligonos < m_uMaxPoligonos )
   	m_uNumPoligonos ++;

   // define poligonos a serem ordenados
   mOrdenarPols ();
}



/* Define poligonos que deverao ser ordenados. Chamado uma
vez no solido temporario antes de iniciar ordenacoes de poligonos.
**********************************************************/
void CSolido::mOrdenarPols ()
{
   // calcula distancia do ponto medio dos poligonos ateh a origem
   for ( UINT16 ucP = 0; ucP < m_uNumPoligonos; ucP ++ )

      // guarda indice
      * ( m_pauP + ucP ) = ucP;
}



/* Atualiza matriz de distancias de poligonos e faz ordenacao
dos poligonos conforme a distancia em Z. Para definir o poligono
cuja distancia serah calculada eh usado o indice que estah na tabela
de indices de poligonos ordenados por proximidade. Apohs ser
calculada a distancia em Z de cada poligono eh feita a atualizacao da
tabela de indices de poligonos ordenados, reordenando-os conforme as
novas distancias obtidas. A ordenacao eh crescente. Usa bubblesort.
O efeito dessa forma de ordenacao eh que apenas os poligonos que
trocaram sua proximidade com a origem em relacao a um outro poligono
eh que terah a sua ordem alterada (se um poligono que estava mais a
frente que outro passou para tras entao ele terah sua ordem alterada).
**********************************************************/
void CSolido::mOrdena ()
{
   // calcula distancia do ponto medio do poligono ateh a origem
   for ( UINT16 ucP = 0; ucP < m_uNumPoligonos; ucP ++ ) {

   	UINT16
      	uNdxProx = * ( m_pauP + ucP );

      // calcula distancia Z do ponto medio
      * ( m_panD + ucP ) = (INT16)
    	                     ( * ( m_panZ + * ( m_pauV1 + uNdxProx ) ) +
                             * ( m_panZ + * ( m_pauV2 + uNdxProx ) ) +
                             * ( m_panZ + * ( m_pauV3 + uNdxProx ) ) +
                             * ( m_panZ + * ( m_pauV4 + uNdxProx ) ) ) >> 2;
   }

   // ordena indices conforme proximidade com a origem, usa bubblesort
	BOOL
		bFim;
   UINT16
   	uPosFim = m_uNumPoligonos;

	// faz ateh estar ordenado (nao haverem trocas por uma varredura)
	do	{

		bFim = TRUE;

		// executa cada interacao varrento a matriz de distancias
		for ( UINT16 ucPos = 0; ucPos < ( uPosFim - 1 ); ucPos ++ )

			// se uma distancia eh maior que a seguinte entao troca
			if ( * ( m_panD + ucPos ) > * ( m_panD + ucPos + 1 ) ) {

				INT16
					nDist = * ( m_panD + ucPos );
            UINT16
            	uNdx  = * ( m_pauP + ucPos );
				* ( m_panD + ucPos ) = * ( m_panD + ucPos + 1 );
           	* ( m_pauP + ucPos ) = * ( m_pauP + ucPos + 1 );
				* ( m_panD + ucPos + 1 ) = nDist;
           	* ( m_pauP + ucPos + 1 ) = uNdx;
				bFim = FALSE;
			}

		uPosFim --;

	} while ( ! bFim );
}



/* return maximo de vertices que o solido pode ter
**********************************************************/
UINT16 CSolido::mRetMaxVertices ()
{
	return m_uMaxVertices;
}



/* return maximo de poligonos que o solido pode ter
**********************************************************/
UINT16 CSolido::mRetMaxPoligonos ()
{
	return m_uMaxPoligonos;
}
